require("RYUKU_StarvingZombies_Vector")
require("RYUKU_StarvingZombies_Bodies")
if isServer() then return end

---------------------------
-- IsoDeadBody.deathTime --
---------------------------

local deathIdx

---@param isoDeadBody zombie.iso.objects.IsoDeadBody
---@return number
local function getDeathTime(isoDeadBody)
    if deathIdx == nil then
        for i = 0, getNumClassFields(isoDeadBody) - 1 do
            if tostring(getClassField(isoDeadBody, i)) == "private float zombie.iso.objects.IsoDeadBody.deathTime" then
                deathIdx = i
                break
            end
        end

        if deathIdx == nil then
            getPlayer():addLineChatElement("Starving Zombies: failed to get deathIdx")
            return getGameTime():getWorldAgeHours()
        end
    end

    return getClassFieldVal(isoDeadBody, getClassField(isoDeadBody, deathIdx))
end

-----------
-- Smell --
-----------

local szVars
local climateMan
local objVec = SZVector:new()
local charVec = SZVector:new()
local charDir
local charDist
local smellDist

---@param isoZombie zombie.characters.IsoZombie
---@param isoObject zombie.iso.IsoObject
local function initSmellVars(isoZombie, isoObject)
    charVec.x = isoZombie:getX()
    charVec.y = isoZombie:getY()
    charVec.z = isoZombie:getZ()
    objVec.x = isoObject:getX()
    objVec.y = isoObject:getY()
    objVec.z = isoObject:getZ()
    charDir = charVec:sub(objVec)
    charDist = charDir:mag()
    charDir:normalise()
end

---@param maxDistance integer
---@param windSpeedMult number
---@param dispersionAngle integer
---@param bodySizeMult number?
---@return boolean
local function canSmell(maxDistance, windSpeedMult, dispersionAngle, bodySizeMult)
    if not climateMan then climateMan = getClimateManager() end
    if charDist > maxDistance then return false end
    if charDist <= smellDist then return true end

    smellDist = smellDist + climateMan:getWindspeedKph() * windSpeedMult
    if bodySizeMult ~= nil then smellDist = smellDist * bodySizeMult end
    if charDist > smellDist then return false end

    if dispersionAngle > 0 then
        local windRads = climateMan:getWindAngleRadians()
        local windDir = SZVector:new()
        windDir.x = math.cos(windRads)
        windDir.y = math.sin(windRads)
        if charDir:dot(windDir) >= 1 - dispersionAngle / 90.0 then return true end
        return false
    end

    return true
end

----------------
-- Body Smell --
----------------

local gameTime
local bodySquare
local zombieSquare

---@param isoZombie zombie.characters.IsoZombie
---@param isoDeadBody zombie.iso.objects.IsoDeadBody
---@return boolean
function szCanSmellBody(isoZombie, isoDeadBody)
    if not szVars then szVars = SandboxVars.StarvingZombies end
    if not gameTime then gameTime = getGameTime() end
    if not climateMan then climateMan = getClimateManager() end

    -- indoor check
    if szVars.BodyIndoorCheck then
        bodySquare = isoDeadBody:getCurrentSquare()
        if bodySquare == nil then return false end

        zombieSquare = isoZombie:getCurrentSquare()
        if zombieSquare == nil then return false end

        if not bodySquare:isOutside() and isoZombie:isOutside() then return false end
        if bodySquare:getBuilding() ~= zombieSquare:getBuilding() then return false end
    end

    -- smell distance
    initSmellVars(isoZombie, isoDeadBody)
    smellDist = szVars.BodyStartDistance
        + (gameTime:getWorldAgeHours() - getDeathTime(isoDeadBody)) * szVars.BodyAgeMultiplier

    if szVars.BodyTemperaturePeak > 0 then
        smellDist = smellDist * math.max(0, climateMan:getTemperature()) / szVars.BodyTemperaturePeak
    end

    local bodyMult = nil
    if szVars.BodySizeImpactsSmell then
        if isoDeadBody:isAnimal() then bodyMult = szBodyMultiplier[isoDeadBody:getModData().AnimalType]
        else bodyMult = szBodyMultiplier.human end
    end

    return canSmell(szVars.BodyMaxDistance, szVars.BodyWindSpeedMultiplier, szVars.BodyDispersionAngle, bodyMult)
end

------------------
-- Player Smell --
------------------

---@param inventory zombie.inventory.ItemContainer
---@param blood number
---@return number
local function getBloodOnItems(inventory, blood)
    local item
    local coveredParts
    local items = inventory:getItems()
    for i = 0, items:size() - 1 do
        item = items:get(i)
        if instanceof(item, "Clothing") then
            coveredParts = BloodClothingType.getCoveredParts(item:getBloodClothingType())
            if coveredParts then
                for x = 0, coveredParts:size() - 1 do
                    blood = blood + item:getBlood(coveredParts:get(x))
                end
            end
        else
            blood = blood + item:getBloodLevel()
        end

        if item:IsInventoryContainer() then
            blood = getBloodOnItems(item:getInventory(), blood)
        end
    end

    return blood
end

---@param isoPlayer zombie.characters.IsoPlayer
local function onPlayerUpdate(isoPlayer)
    if not szVars then szVars = SandboxVars.StarvingZombies end

    local time = getTimeInMillis()
    local modData = isoPlayer:getModData()
    if modData.szUpdate ~= nil and time < modData.szUpdate then return end
    modData.szUpdate = time + 1000

    local smell = szVars.PlayerStartDistance
    local bloodMult = szVars.PlayerBloodMultiplier
    local visual = isoPlayer:getHumanVisual()
    for i = 0, BloodBodyPartType.MAX:index() - 1 do
        smell = smell + visual:getBlood(BloodBodyPartType.FromIndex(i)) * bloodMult
    end

    modData.szSmell = smell + getBloodOnItems(isoPlayer:getInventory(), 0) * bloodMult
end

Events.OnPlayerUpdate.Add(onPlayerUpdate)

---@param isoZombie zombie.characters.IsoZombie
---@param isoPlayer zombie.characters.IsoPlayer
---@return boolean
function szCanSmellPlayer(isoZombie, isoPlayer)
    if not szVars then szVars = SandboxVars.StarvingZombies end

    initSmellVars(isoZombie, isoPlayer)
    smellDist = isoPlayer:getModData().szSmell
    if smellDist == nil then return false end
    return canSmell(szVars.PlayerMaxDistance, szVars.PlayerWindSpeedMultiplier, szVars.PlayerDispersionAngle)
end
